use std::collections::HashMap;

use mongod::{assert_reference_of, reference_of, Reference, Validate};

#[derive(Debug, Clone, Serialize, Deserialize, Model, Referencable)]
pub struct FlowInfo {
    #[serde(default)]
    pub _id: String,
    pub name: String,
    #[serde(default)]
    pub depends: Vec<String>,
    pub next: Option<String>,
    pub produces: Option<Vec<String>>,
}

impl Validate for FlowInfo {
    async fn validate(&self) -> Result<(), String> {
        Ok(())
    }
}

impl FlowInfo {
    pub async fn add(&mut self, id: &str) {
        self._id = id.to_string();
        self.insert_overwrite().await.unwrap();
    }
}

impl ToAPI for FlowInfo {
    async fn api(&self) -> serde_json::Value {
        json!({
            "id": self._id,
            "name": self.name,
            "depends": self.depends,
            "next": self.next,
            "produces": self.produces
        })
    }
}

use mongod::{
    derive::{Model, Referencable},
    Model, Referencable, ToAPI,
};
use mongodb::bson::doc;
use serde::{Deserialize, Serialize};
use serde_json::json;

use crate::item::Item;
use crate::routes::item::{item_does_not_exist_error, variant_does_not_exist_error};
use crate::routes::{api_error, ApiError};
use crate::transaction::{Price, Transaction};

#[derive(Debug, Clone, Serialize, Deserialize, PartialEq, Eq)]
pub struct DoneInfo {
    /// Timestamp when the flow was ended
    pub ended: i64,
    /// The flow succedding this one
    pub next: Option<Reference>,
    /// Transactions this flow produced
    pub produced: Option<Vec<Reference>>,
}

impl DoneInfo {
    pub fn new(next: Option<Reference>) -> Self {
        Self {
            ended: chrono::Utc::now().timestamp(),
            next,
            produced: None,
        }
    }

    pub fn api(&self) -> serde_json::Value {
        json!({
            "ended": self.ended,
            "next": self.next.as_ref().map(|x| x.id()),
            "produced": self.produced.as_ref().map(|x| x.iter().map(|t| t.id()).collect::<Vec<_>>()),
        })
    }
}

/// A production flow
#[derive(Debug, Clone, Serialize, Deserialize, Model, Referencable)]
pub struct Flow {
    /// ID
    pub _id: String,
    /// Tiemstamp when the flow was started
    pub started: i64,
    /// Kind of flow; ID of the describing JSON
    pub kind: Reference,
    /// Input transactions
    pub input: Option<Vec<Reference>>,
    /// Information when a flow is done
    pub done: Option<DoneInfo>,
}

impl Validate for Flow {
    async fn validate(&self) -> Result<(), String> {
        assert_reference_of!(self.kind, FlowInfo);

        if let Some(input) = &self.input {
            for t in input {
                assert_reference_of!(t, Transaction);
            }
        }

        if let Some(done) = &self.done {
            if let Some(next) = &done.next {
                assert_reference_of!(next, Flow);
            }
            if let Some(produced) = &done.produced {
                for prod in produced {
                    assert_reference_of!(prod, Transaction);
                }
            }
        }

        Ok(())
    }
}

impl ToAPI for Flow {
    async fn api(&self) -> serde_json::Value {
        json!({
            "id": self._id,
            "started": self.started,
            "kind": self.kind.id(),
            "input": self.input.as_ref().map(|x| x.iter().map(|t| t.id()).collect::<Vec<_>>()),
            "done": self.done.as_ref().map(|x| x.api())
        })
    }
}

impl Flow {
    pub async fn create(kind: &str, input: Option<Vec<Reference>>) -> Self {
        let f = Self {
            _id: uuid::Uuid::new_v4().to_string(),
            started: chrono::Utc::now().timestamp(),
            kind: reference_of!(FlowInfo, kind).unwrap(),
            input: input,
            done: None,
        };

        f.insert().await.unwrap();

        f
    }

    pub async fn end(self) -> Self {
        self.change()
            .done(Some(DoneInfo::new(None)))
            .update()
            .await
            .unwrap()
    }

    pub async fn end_with_produce(
        self,
        produced: &[HashMap<String, u64>],
    ) -> Result<HashMap<String, Vec<String>>, ApiError> {
        let mut ret = HashMap::new();
        let mut t_create = Vec::new();
        let mut produced_ref = Vec::with_capacity(ret.len());

        for prod in produced {
            for (item_variant, amount) in prod {
                let (item, variant) = item_variant
                    .split_once("::")
                    .ok_or_else(|| api_error("Bad Request"))?;
                let t = Item::get(item)
                    .await
                    .ok_or_else(item_does_not_exist_error)?
                    .variant(variant)
                    .ok_or_else(variant_does_not_exist_error)?;

                t_create.push((t, amount));
            }
        }

        for (item, amount) in t_create {
            for _ in 0..*amount {
                let t = item
                    .supply(
                        Price::zero(),
                        Some(&format!("flow::{}::{}", self.kind.id(), self._id)),
                        None,
                        None,
                    )
                    .await;
                ret.entry(item.item_variant_id().clone())
                    .or_insert(Vec::new())
                    .push(t._id.clone());
                produced_ref.push(t.reference());
            }
        }

        self.change()
            .done(Some(DoneInfo {
                ended: chrono::Utc::now().timestamp(),
                next: None,
                produced: Some(produced_ref),
            }))
            .update()
            .await
            .unwrap();

        Ok(ret)
    }

    pub async fn continue_next(self, next_flow: &Flow) -> Self {
        self.change()
            .done(Some(DoneInfo::new(Some(next_flow.reference()))))
            .update()
            .await
            .unwrap()
    }
}

/// A note for a Flow
#[derive(Debug, Clone, Serialize, Deserialize, Model, Referencable)]
pub struct FlowNote {
    /// ID
    pub _id: String,
    /// Tiemstamp when the note was created
    pub timestamp: i64,
    /// Note Content
    pub content: String,
    /// Associated flow
    pub on_flow: Reference,
}

impl FlowNote {
    pub async fn create(content: &str, flow: Reference) -> Self {
        let s = Self {
            _id: uuid::Uuid::new_v4().to_string(),
            timestamp: chrono::Utc::now().timestamp(),
            content: content.to_string(),
            on_flow: flow,
        };

        s.insert().await.unwrap();

        s
    }
}

impl ToAPI for FlowNote {
    async fn api(&self) -> serde_json::Value {
        json!({
            "uuid": self._id,
            "timestamp": self.timestamp,
            "content": self.content,
            "on_flow": self.on_flow.id()
        })
    }
}

impl Validate for FlowNote {
    async fn validate(&self) -> Result<(), String> {
        assert_reference_of!(self.on_flow, Flow);

        Ok(())
    }
}
