mod demand;
pub mod error;
mod location;
mod stat;
mod supply;

pub use demand::*;
pub use error::*;
pub use location::*;
use mongod::reference_of;
use mongod::Model;
use mongod::ToAPI;
use rocket::post;
use rocket::serde::json::Json;
use serde::Deserialize;
use serde::Serialize;
pub use stat::*;
pub use supply::*;

use rocket::get;
use rocket::State;
use serde_json::json;

use crate::check_auth;
use crate::config::Config;
use crate::db::get_items_without_min_satisfied;
use crate::db::ItemDB;
use crate::location::Location;
use crate::routes::Token;
use crate::transaction::Transaction;

use super::api_error;
use crate::routes::FallibleApiResponse;

/// Returns a JSON response with all items in the database.
#[get("/items")]
pub fn get_items_route(itemdb: &State<ItemDB>, t: Token, c: &State<Config>) -> FallibleApiResponse {
    check_auth!(t, c);

    let mut items = Vec::new();
    for item in itemdb.items() {
        items.push(itemdb.get_item(&item).unwrap().api_json());
    }

    Ok(json!({"items": items}))
}

#[get("/items/stat")]
pub async fn item_stat_route(
    itemdb: &State<ItemDB>,
    t: Token,
    c: &State<Config>,
) -> FallibleApiResponse {
    check_auth!(t, c);

    let items = itemdb.items();
    let item_count = items.len();
    let mut transaction_count = 0;
    let mut total_price = 0.0;

    for item in items {
        for var in itemdb.get_item(&item).unwrap().variants.keys() {
            let item_var = itemdb.get_item(&item).unwrap().variant(var).unwrap();
            for t in item_var.inventory().await {
                transaction_count += 1;
                total_price += t.price.value;
            }
        }
    }

    Ok(json!({
        "item_count": item_count,
        "total_transactions": transaction_count,
        "total_price": total_price
    }))
}

/// Return an API Response for an `Item`
#[get("/item/<item_id>")]
pub fn item_route(
    item_id: &str,
    itemdb: &State<ItemDB>,
    t: Token,
    c: &State<Config>,
) -> FallibleApiResponse {
    check_auth!(t, c);

    let item = itemdb
        .get_item(item_id)
        .ok_or_else(item_does_not_exist_error)?;
    Ok(item.api_json())
}

/// Returns all variants of an Item
#[get("/item/<item_id>/variants")]
pub fn item_variants_page(
    item_id: &str,
    itemdb: &State<ItemDB>,
    t: Token,
    c: &State<Config>,
) -> FallibleApiResponse {
    check_auth!(t, c);

    let item = itemdb
        .get_item(item_id)
        .ok_or_else(item_does_not_exist_error)?;
    let variants = item.get_variants();

    Ok(json!({
        "item": item_id,
        "variants": variants
    }))
}

/// Returns an API Response for a `Transaction`
#[get("/transaction/<transaction>")]
pub async fn transaction_route(
    transaction: &str,
    t: Token,
    c: &State<Config>,
) -> FallibleApiResponse {
    check_auth!(t, c);

    let t = Transaction::get(transaction)
        .await
        .ok_or_else(|| api_error("No transaction with this UUID"))?;
    Ok(t.api().await)
}

/// Returns unique values for a field
#[get("/item/<item_id>/<variant_id>/unique?<field>")]
pub async fn unique_field_route(
    item_id: &str,
    variant_id: &str,
    field: &str,
    itemdb: &State<ItemDB>,
    t: Token,
    c: &State<Config>,
) -> FallibleApiResponse {
    check_auth!(t, c);

    let variant = itemdb
        .get_item(item_id)
        .ok_or_else(item_does_not_exist_error)?
        .variant(variant_id)
        .ok_or_else(variant_does_not_exist_error)?;

    match field {
        "origin" => Ok(json!(variant.get_unique_origins().await)),
        "destination" => Ok(json!(variant.get_unique_destinations().await)),
        _ => Err(api_error("Unknown field")),
    }
}

#[get("/items/expired")]
pub async fn expired_items_route(t: Token, c: &State<Config>) -> FallibleApiResponse {
    check_auth!(t, c);

    let t = Transaction::active_expired().await;

    Ok(json!(t))
}

#[get("/items/min")]
pub async fn min_items_route(
    itemdb: &State<ItemDB>,
    t: Token,
    c: &State<Config>,
) -> FallibleApiResponse {
    check_auth!(t, c);

    let t: Vec<_> = get_items_without_min_satisfied(itemdb)
        .await
        .into_iter()
        .map(|x| {
            json!({
                "item_variant": x.0,
                "need": x.1
            }
            )
        })
        .collect();

    Ok(json!(t))
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct MoveTransaction {
    pub to: String,
}

#[post("/transaction/<id>/move", data = "<form>")]
pub async fn move_transaction_route(id: &str, form: Json<MoveTransaction>) -> FallibleApiResponse {
    let new_loc = &form.to;
    Transaction::get(id)
        .await
        .ok_or_else(|| api_error("No such transaction"))?
        .change()
        .location(Some(
            reference_of!(Location, new_loc).ok_or_else(|| api_error("No such location"))?,
        ))
        .update()
        .await
        .unwrap();

    Ok(json!({"ok": 1 }))
}
